""" Markers and controls

See also https://docs.mapbox.com/mapbox-gl-js/api/markers/
"""

from enum import Enum
from typing import Literal, Optional, Union

from pydantic import Field

from ._core import MapLibreBaseModel


class PopupOptions(MapLibreBaseModel):
    """Popup options"""

    anchor: str = None
    close_button: bool = Field(False, serialization_alias="closeButton")
    close_on_click: bool = Field(None, serialization_alias="closeOnClick")
    close_on_move: bool = Field(None, serialization_alias="closeOnMove")
    max_width: int = Field(None, serialization_alias="maxWidth")
    offset: Union[int, list, dict] = None


class Popup(MapLibreBaseModel):
    """Popup

    Attributes:
        text: The Text of the popup.
        options (PopupOptions | dict): Popup options.
    """

    text: str
    options: Union[PopupOptions, dict] = {}


class MarkerOptions(MapLibreBaseModel):
    """Marker options"""

    anchor: str = None
    color: str = None
    draggable: bool = None
    offset: Union[tuple, list] = None
    pitch_alignment: str = Field(None, serialization_alias="pitchAlignment")
    rotation: int = None
    rotation_alignment: str = Field(None, serialization_alias="rotationAlignment")
    scale: int = None


class Marker(MapLibreBaseModel):
    """Marker

    Attributes:
        lng_lat (tuple |list): **Required.** The longitude and latitude of the marker.
        popup (Popup | dict): The Popup that is displayed when a user clicks on the marker.
        options (MarkerOptions | dict): Marker options.
    """

    lng_lat: Union[tuple, list] = Field(None, serialization_alias="lngLat")
    popup: Union[Popup, dict] = None
    options: Union[MarkerOptions, dict] = {}


class ControlType(Enum):
    NAVIGATION = "NavigationControl"
    SCALE = "ScaleControl"
    FULLSCREEN = "FullscreenControl"
    GEOLOCATE = "GeolocateControl"
    ATTRIBUTION = "AttributionControl"


class ControlPosition(Enum):
    """Control position

    Attributes:
        TOP_LEFT: top-left
        TOP_RIGHT: top-right
        BOTTOM_LEFT: bottom-left
        BOTTOM_RIGHT: bottom-right
    """

    TOP_LEFT = "top-left"
    TOP_RIGHT = "top-right"
    BOTTOM_LEFT = "bottom-left"
    BOTTOM_RIGHT = "bottom-right"


class Control(MapLibreBaseModel):
    position: Union[ControlPosition, str] = Field(
        ControlPosition.TOP_RIGHT, exclude=True
    )

    @property
    def type(self):
        return self.__class__.__name__


class AttributionControl(Control):
    """Attribution control"""

    # _name: str = ControlType.ATTRIBUTION.value
    compact: bool = None
    custom_attribution: Union[str, list] = Field(
        None, serialization_alias="customAttribution"
    )


class FullscreenControl(Control):
    """Fullscreen control

    Examples:
        >>> from maplibre import Map
        >>> from maplibre.controls import FullscreenControl, ControlPosition

        >>> m = Map()
        >>> m.add_control(FullscreenControl(), ControlPosition.BOTTOM_LEFT)
    """

    # _name: str = ControlType.FULLSCREEN.value
    pass


class GeolocateControl(Control):
    """Geolocate control"""

    # _name: str = ControlType.GEOLOCATE.value
    position_options: dict = Field(None, serialization_alias="positionOptions")
    show_accuracy_circle: bool = Field(True, serialization_alias="showAccuracyCircle")
    show_user_heading: bool = Field(False, serialization_alias="showUserHeading")
    show_user_location: bool = Field(True, serialization_alias="showUserLocation")
    track_user_location: bool = Field(False, serialization_alias="trackUserLocation")


class NavigationControl(Control):
    """Navigation control"""

    # _name: str = ControlType.NAVIGATION.value
    show_compass: bool = Field(True, serialization_alias="showCompass")
    show_zoom: bool = Field(True, serialization_alias="showZoom")
    visualize_pitch: bool = Field(False, serialization_alias="visualizePitch")


class ScaleUnit(Enum):
    IMPERIAL = "imperial"
    METRIC = "metric"
    NAUTICAL = "nautical"


class ScaleControl(Control):
    """Scale control"""

    max_width: int = Field(None, serialization_alias="maxWidth")
    unit: Literal["imperial", "metric", "nautical"] = "metric"


# -------------------------
# Custom controls
# -------------------------
class LayerSwitcherControl(Control):
    """Layer switcher control

    Attributes:
        layer_ids (list): A list of layer ids to be shown in the layer switcher control.
        theme (Literal["default", "simple"]): The theme of the layer switcher control.
        css_text (str): Optional inline style declaration of the control.
    """

    layer_ids: list = Field([], serialization_alias="layerIds")
    theme: Literal["default", "simple"] = "default"
    css_text: str = Field(None, serialization_alias="cssText")


class InfoBoxControl(Control):
    """InfoBox control

    Attributes:
        content (str): Content (HTML or plain text) to be displayed in the info box.
        css_text (str): Optional inline style declaration of the control.
    """

    content: str
    css_text: str = Field(None, serialization_alias="cssText")



from __future__ import annotations

from enum import Enum
from typing import Optional, Union
from uuid import uuid4

from pydantic import Field, field_validator

from ._core import MapLibreBaseModel
from ._utils import fix_keys
from .sources import SimpleFeatures, Source
from .utils import get_bounds

try:
    from geopandas import GeoDataFrame
except ImportError:
    GeoDataFrame = None


class LayerType(Enum):
    """Rendering type of layer

    Attributes:
        CIRCLE: A filled circle.
        FILL: A filled polygon with an optional stroked border.
        FILL_EXTRUSION: An extruded polygon.
        LINE: A stroked line.
        SYMBOL: An icon or a text label.
        RASTER: Raster map textures such as satellite imagery.
        HEATMAP: A heatmap.
        HILLSHADE: A Client-side hillshading visualization based on DEM data.
        BACKGROUND: A background color or pattern.
    """

    CIRCLE = "circle"
    FILL = "fill"
    FILL_EXTRUSION = "fill-extrusion"
    LINE = "line"
    SYMBOL = "symbol"
    RASTER = "raster"
    HEATMAP = "heatmap"
    HILLSHADE = "hillshade"
    BACKGROUND = "background"


class Layer(MapLibreBaseModel):
    """Layer properties

    Notes:
        See [layers](https://maplibre.org/maplibre-style-spec/layers/) for more details on the
        `paint` and `layout` properties of the layers.

    Attributes:
        id (str): **Required.** The unique ID of the layer. Defaults to `str(uuid4())`.
        type (str | LayerType): **Required.** The type of the layer.
        filter (list): A filter expression that is applied to the source of the layer.
        layout (dict): The layout properties for the layer.
        max_zoom (int): The maximum zoom level for the layer.
        min_zoom (int): The minimum zoom level for the layer.
        paint (dict): The paint properties for the layer.
        source (str | Source | GeoDataFrame): The name (unique ID) of a source, a source object or a GeoDataFrame
            to be used for the layer.
        source_layer (str): The layer to use from a vector tile source.

    Examples:
        >>> from maplibre.layer import Layer, LayerType

        >>> layer = Layer(id="test-layer", type=LayerType.CIRCLE, source="test-source")
    """

    id: Optional[str] = Field(default_factory=lambda: str(uuid4()))
    type: LayerType
    filter: Optional[list] = None
    layout: Optional[dict] = None
    max_zoom: Optional[int] = Field(None, serialization_alias="maxzoom")
    metadata: Optional[dict] = None
    min_zoom: Optional[int] = Field(None, serialization_alias="minzoom")
    paint: Optional[dict] = None
    source: Union[str, Source, dict, GeoDataFrame, None] = None
    source_layer: Optional[str] = Field(None, serialization_alias="source-layer")

    # TODO: Use model_post_init and set bounds if source is a GeoDataFrame
    @field_validator("source")
    def validate_source(cls, v):
        if GeoDataFrame is not None and isinstance(v, GeoDataFrame):
            return SimpleFeatures(v).to_source().to_dict()

        if isinstance(v, Source):
            return v.to_dict()

        return v

    @field_validator("paint", "layout")
    def fix_paint(cls, v):
        if isinstance(v, dict):
            return fix_keys(v)

        return v

    @property
    def bounds(self) -> tuple | None:
        try:
            bounds = get_bounds(self.source["data"])
        except Exception as e:
            # print(e)
            bounds = None

        return bounds

    def set_paint_props(self, **props) -> Layer:
        if self.paint is None:
            self.paint = dict()

        self.paint = self.paint | fix_keys(props)
        return self

    def set_layout_props(self, **props) -> Layer:
        if self.layout is None:
            self.layout = dict()

        self.paint = self.paint | fix_keys(props)
        return self



from htmltools import Tag
from shiny.render.renderer import Renderer

from .map import Map
from .ui import output_maplibregl


class MapLibreRenderer(Renderer[Map]):
    def auto_output_ui(self) -> Tag:
        return output_maplibregl(self.output_id, height=600)

    async def transform(self, value: Map) -> dict:
        return {"mapData": value.to_dict()}



from __future__ import annotations

from pydantic import Field

from ._core import MapLibreBaseModel


class MapboxDrawControls(MapLibreBaseModel):
    """MapboxDraw controls"""

    point: bool = False
    line_string: bool = False
    polygon: bool = False
    trash: bool = False
    combine_features: bool = False
    uncombine_features: bool = False


class MapboxDrawOptions(MapLibreBaseModel):
    """MapboxDraw Options"""

    display_controls_default: bool = Field(
        True, serialization_alias="displayControlsDefault"
    )
    controls: MapboxDrawControls = None
    box_select: bool = Field(True, serialization_alias="boxSelect")



from __future__ import annotations

from enum import Enum
from typing import Optional, Union
from uuid import uuid4

from pydantic import Field, computed_field, field_validator

from ._core import MapLibreBaseModel
from .utils import geopandas_to_geojson

try:
    from geopandas import GeoDataFrame, read_file
except ImportError:
    GeoDataFrame, read_file = None, None

CRS = "EPSG:4326"


class SourceType(Enum):
    """Source types"""

    RASTER = "raster"
    VECTOR = "vector"
    RASTER_DEM = "raster-dem"
    GEOJSON = "geojson"
    IMAGE = "image"
    VIDEO = "video"


class Source(MapLibreBaseModel):
    pass


class GeoJSONSource(Source):
    """GeoJSON Source

    Examples:
        >>> from maplibre.sources import GeoJSONSource

        >>> geojson = "https://docs.mapbox.com/mapbox-gl-js/assets/earthquakes.geojson"
        >>> source = GeoJSONSource(data=geojson)
    """

    data: Union[str, dict]
    attribution: Optional[str] = None
    buffer: Optional[int] = None
    cluster: Optional[bool] = None
    cluster_max_zoom: Optional[int] = Field(None, serialization_alias="clusterMaxZoom")
    cluster_min_points: Optional[int] = Field(
        None, serialization_alias="clusterMinPoints"
    )
    cluster_properties: Optional[dict] = Field(
        None, serialization_alias="clusterProperties"
    )
    cluster_radius: Optional[int] = Field(None, serialization_alias="clusterRadius")
    filter: Optional[list] = None
    generate_id: Optional[bool] = Field(None, serialization_alias="generateId")
    line_metrics: Optional[bool] = Field(None, serialization_alias="lineMetrics")
    min_zoom: Optional[int] = Field(None, serialization_alias="minzoom")
    max_zoom: Optional[int] = Field(None, serialization_alias="maxzoom")
    promote_id: Union[str, dict, None] = Field(None, serialization_alias="promoteId")
    tolerance: Optional[float] = None

    @computed_field
    @property
    def type(self) -> str:
        return SourceType.GEOJSON.value


class RasterTileSource(Source):
    """Raster tile source

    Examples:
        >>> from maplibre.sources import RasterTileSource

        >>> raster_tile_source = RasterTileSource(
        ...     tiles=["https://tile.openstreetmap.org/{z}/{x}/{y}.png"],
        ...     tile_size=256,
        ...     min_zoom=0,
        ...     max_zoom=19,
        ... )
    """

    attribution: str = None
    bounds: tuple = None
    max_zoom: int = Field(None, serialization_alias="maxzoom")
    min_zoom: int = Field(None, serialization_alias="minzoom")
    scheme: str = None
    tile_size: int = Field(None, serialization_alias="tileSize")
    tiles: Union[tuple, list] = None
    url: str = None
    volatile: bool = None

    @computed_field
    @property
    def type(self) -> str:
        return SourceType.RASTER.value


class VectorTileSource(Source):
    """Vector tile source

    Examples:
        >>> from maplibre.sources import VectorTileSource
        >>> from maplibre import LayerType, Layer

        >>> vector_tile_source = VectorTileSource(
        ...     tiles=["https://demotiles.maplibre.org/tiles/{z}/{x}/{y}.pbf"],
        ...     min_zoom=0,
        ...     max_zoom=6,
        ... )

        >>> layer = Layer(
        ...     type=LayerType.LINE,
        ...     id="countries",
        ...     source=vector_tile_source,
        ...     source_layer="countries",
        ...     paint={"fill-color": "lightgreen", "fill-outline-color": "black"},
        ... )
    """

    attribution: str = None
    bounds: tuple = None
    max_zoom: int = Field(None, serialization_alias="maxzoom")
    min_zoom: int = Field(None, serialization_alias="minzoom")
    scheme: str = None
    tiles: Union[tuple, list] = None
    url: str = None
    volatile: bool = None

    @computed_field
    @property
    def type(self) -> str:
        return SourceType.VECTOR.value


class SimpleFeatures(object):
    def __init__(self, data: GeoDataFrame | str, source_id: str = None):
        self._path = data if isinstance(data, str) else None
        if self._path is not None:
            data = read_file(self._path)

        if str(data.crs) != CRS:
            data = data.to_crs(CRS)

        self._data = data
        self._source_id = source_id or str(uuid4())

    @property
    def path(self):
        return self._path

    @property
    def crs(self):
        return self._data.crs

    @property
    def bounds(self) -> tuple:
        return self._data.total_bounds

    @property
    def source_id(self) -> str:
        return self._source_id

    @property
    def data(self) -> GeoDataFrame:
        return self._data

    def to_source(self, **kwargs) -> GeoJSONSource:
        kwargs["data"] = geopandas_to_geojson(self._data)
        return GeoJSONSource(**kwargs)

    def to_sources_dict(self, **kwargs) -> dict:
        return {self.source_id: self.to_source(**kwargs)}



import os
from typing import Optional

from pydantic import BaseModel


class Options(BaseModel):
    cmap: Optional[str] = "viridis"
    fallback_color: Optional[str] = "#000000"

    fill_color: Optional[str] = "#3B9AB2"
    fill_opacity: Optional[float] = 0.5
    fill_outline_color: Optional[str] = "#FFFFFF"

    fill_extrusion_color: Optional[str] = "#3B9AB2"
    fill_extrusion_opacity: Optional[float] = 1.0

    line_color: Optional[str] = "#F21A00"
    line_opacity: Optional[float] = 1.0
    line_width: Optional[float] = 1.0

    circle_color: Optional[str] = "#EBCC2A"
    circle_opacity: Optional[float] = 1.0
    circle_radius: Optional[float] = 5
    circle_stroke_color: Optional[str] = "#EBCC2A"

    maptiler_api_key_env_var: Optional[str] = "MAPTILER_API_KEY"

    layer_types: Optional[dict] = dict(
        Polygon="fill", LineString="line", Point="circle"
    )

    @property
    def maptiler_api_key(self) -> str:
        return os.environ.get(self.maptiler_api_key_env_var)

    @property
    def paint_props(self) -> dict:
        return dict(
            fill={
                "fill-color": self.fill_color,
                "fill-opacity": self.fill_opacity,
                "fill-outline-color": self.fill_outline_color,
            },
            fill_extrusion={
                "fill-extrusion-color": self.fill_extrusion_color,
                "fill-extrusion-opacity": self.fill_extrusion_opacity,
            },
            line={
                "line-color": self.line_color,
                "line-opacity": self.line_opacity,
                "line-width": self.line_width,
            },
            circle={
                "circle-color": self.circle_color,
                "circle-opacity": self.circle_opacity,
                "circle-radius": self.circle_radius,
                "circle-stroke-color": self.circle_stroke_color,
            },
        )

    def set_maptiler_api_key(self, api_key: str) -> None:
        os.environ[self.maptiler_api_key_env_var] = api_key


options = Options()



### TODO: move to 'shiny.py'

from __future__ import annotations

from htmltools import HTMLDependency, Tag
from shiny import ui
from shiny.module import resolve_id

from ._constants import __version__, _shiny_output_class

MAPLIBREGL_VERSION = "3.6.2"

maplibregl_dep = HTMLDependency(
    "maplibregl",
    version=MAPLIBREGL_VERSION,
    source={"package": "maplibre", "subdir": "srcjs"},
    script={"src": "maplibre-gl.js", "type": "module"},
    stylesheet={"href": "maplibre-gl.css"},
    all_files=False,
)

pymaplibregl_dep = HTMLDependency(
    "pymaplibregl",
    version=__version__,
    source={"package": "maplibre", "subdir": "srcjs"},
    script={"src": "pywidget.js", "type": "module"},
    stylesheet={"href": "pywidget.css"},
    all_files=False,
)


def output_maplibregl(id_: str, height: [int | str] = 200) -> Tag:
    if isinstance(height, int):
        height = f"{height}px"

    return ui.div(
        maplibregl_dep,
        pymaplibregl_dep,
        # Use resolve_id so that our component will work in a module
        id=resolve_id(id_),
        class_=_shiny_output_class,
        style=f"height: {height}",
    )


DECKGL_VERSION = "9.0.16"
H3_VERSION = "4.1.0"

h3_dep = HTMLDependency(
    name="h3",
    version=H3_VERSION,
    source={"href": f"https://unpkg.com/h3-js@{H3_VERSION}/dist/"},
    script={"src": "h3-js.umd.js"},
)


def use_h3() -> Tag:
    return ui.div(h3_dep)


deckgl_dep = HTMLDependency(
    name="deckgl",
    version=DECKGL_VERSION,
    source={"href": f"https://unpkg.com/deck.gl@{DECKGL_VERSION}/"},
    script={"src": "dist.min.js", "type": "module"},
)


deckgl_json_dep = HTMLDependency(
    name="deckgljson",
    version=DECKGL_VERSION,
    source={"href": f"https://unpkg.com/@deck.gl/json@{DECKGL_VERSION}/"},
    script={"src": "dist.min.js", "type": "module"},
)


def use_deckgl() -> Tag:
    return ui.div(deckgl_dep, deckgl_json_dep)


MAPBOXGL_DRAW_VERSION = "1.4.3"

mapboxgl_draw_dep = HTMLDependency(
    name="mapbox-gl-draw-plugin",
    version=MAPBOXGL_DRAW_VERSION,
    source={
        "href": f"https://www.unpkg.com/@mapbox/mapbox-gl-draw@{MAPBOXGL_DRAW_VERSION}/dist/"
    },
    script={"src": "mapbox-gl-draw.js", "type": "module"},
    stylesheet={"href": "mapbox-gl-draw.css"},
)


def use_mapboxgl_draw() -> Tag:
    return ui.div(mapboxgl_draw_dep)



from __future__ import annotations

import gzip
import json
from typing import Optional

from pmtiles.reader import MemorySource, MmapSource
from pmtiles.tile import Compression, TileType, deserialize_header
from pydantic import BaseModel

from .basemaps import construct_basemap_style
from .layer import Layer, LayerType
from .sources import RasterTileSource, VectorTileSource

try:
    import requests
except ImportError as e:
    print(e)
    requests = None

# See https://github.com/protomaps/PMTiles/blob/main/spec/v3/spec.md for specs

PMTILES_HEADER_OFFSET = 0
PMTILES_HEADER_LENGTH = 127


class DemoPMTiles(object):
    data_source_coop_openstreetmap = (
        "https://data.source.coop/protomaps/openstreetmap/tiles/v3.pmtiles"
    )
    pmtiles_io_stamen = "https://pmtiles.io/stamen_toner(raster)CC-BY+ODbL_z3.pmtiles"
    r2_public_protomaps_com_us_zcta = "https://r2-public.protomaps.com/protomaps-sample-datasets/cb_2018_us_zcta510_500k.pmtiles"
    pmtiles_io_ugs_mt_whitney = (
        "https://pmtiles.io/usgs-mt-whitney-8-15-webp-512.pmtiles"
    )
    pmtiles_io_vector_firenze_base_layer = (
        "https://pmtiles.io/protomaps(vector)ODbL_firenze.pmtiles"
    )


class PMTilesHeader(BaseModel):
    version: int
    metadata_offset: int
    metadata_length: int
    min_lon_e7: int
    min_lat_e7: int
    max_lon_e7: int
    max_lat_e7: int
    min_zoom: int
    max_zoom: int
    tile_type: TileType
    tile_data_offset: int
    tile_data_length: int
    addressed_tiles_count: int
    tile_entries_count: int
    tile_contents_count: int
    clustered: bool
    center_zoom: int
    center_lon_e7: int
    center_lat_e7: int

    @property
    def center(self):
        return tuple([v / 1e7 for v in [self.center_lon_e7, self.center_lat_e7]])

    @property
    def bounds(self):
        return tuple(
            [
                v / 1e7
                for v in [
                    self.min_lon_e7,
                    self.min_lat_e7,
                    self.max_lon_e7,
                    self.max_lat_e7,
                ]
            ]
        )


class PMTilesMetadata(BaseModel):
    # bounds: tuple
    name: Optional[str] = None
    description: Optional[str] = None
    attribution: Optional[str] = None
    type: Optional[str] = None
    version: Optional[str] = None
    vector_layers: Optional[list] = None

    @property
    def layer_ids(self) -> list:
        return [vector_layer["id"] for vector_layer in self.vector_layers]


def range_request(path: str, offset: int, length: int) -> requests.Response:
    headers = {"Range": f"bytes={offset}-{offset+length}"}
    return requests.get(path, headers=headers)


def get_pmtiles_header(path: str) -> dict:
    if path.startswith("http"):
        response = range_request(path, PMTILES_HEADER_OFFSET, PMTILES_HEADER_LENGTH)
        return deserialize_header(response.content)

    with open(path, "rb") as f:
        get_bytes = MmapSource(f)
        header = deserialize_header(
            get_bytes(PMTILES_HEADER_OFFSET, PMTILES_HEADER_LENGTH)
        )

    return header


def get_pmtiles_metadata(path: str) -> tuple:
    header = get_pmtiles_header(path)
    if path.startswith("http"):
        response = range_request(
            path, header["metadata_offset"], header["metadata_length"]
        )
        get_bytes = MemorySource(response.content)
        metadata = get_bytes(0, header["metadata_length"])
    else:
        with open(path, "rb") as f:
            get_bytes = MmapSource(f)
            metadata = get_bytes(header["metadata_offset"], header["metadata_length"])

    if header["internal_compression"] == Compression.GZIP:
        metadata = gzip.decompress(metadata)

    return header, json.loads(metadata)


class PMTiles(object):
    def __init__(self, path: str):
        self.path = path
        try:
            self._header, self._metadata = get_pmtiles_metadata(path)
        except Exception as e:
            print(e)
            self._header, self._metadata = dict(), dict()

    @property
    def header(self) -> PMTilesHeader:
        return PMTilesHeader(**self._header)

    @property
    def metadata(self) -> PMTilesMetadata:
        return PMTilesMetadata(**self._metadata)

    @property
    def protocol_url(self) -> str:
        return f"pmtiles://{self.path}"

    def layers(self, layer_ids: list = None) -> list:
        pass

    def to_source(self, **kwargs):
        if self.header.tile_type in [TileType.PNG, TileType.JPEG, TileType.WEBP]:
            source = RasterTileSource(url=self.protocol_url)
            if self.metadata.attribution:
                source.attribution = self.metadata.attribution

            return source

        elif self.header.tile_type == TileType.MVT:
            source = VectorTileSource(url=self.protocol_url)
            if self.metadata.attribution:
                source.attribution = self.metadata.attribution

            return source

        return

    def to_basemap_style(self, layer_styles: list) -> dict:
        source_id = self.metadata.name or "pmtiles"
        default_opacity = 1.0
        # Simple layer specs / props
        # layers: [layer_id (source_layer), layer_type, color, opacity]
        layers = []
        for layer_style in layer_styles:
            if len(layer_style) == 3:
                layer_style.append(default_opacity)

            layer_id, layer_type, color, opacity = layer_style
            layer_type = LayerType(layer_type).value
            layers.append(
                Layer(
                    id=layer_id,
                    source=source_id,
                    source_layer=layer_id,
                    type=layer_type,
                    paint={
                        f"{layer_type}-color": color,
                        f"{layer_type}-opacity": opacity,
                    },
                )
            )

        return construct_basemap_style(
            sources={source_id: self.to_source()}, layers=layers
        )



html_template = """<!DOCTYPE html>
<html lang="en">
<meta charset="UTF-8">
<title>{{ title|default('My MapLibre Map')}}</title>
<meta name="viewport" content="width=device-width,initial-scale=1">
<!--
<script src="https://unpkg.com/maplibre-gl/dist/maplibre-gl.js"></script>
<link rel="stylesheet" href="https://unpkg.com/maplibre-gl/dist/maplibre-gl.css"/>
-->
<script src="https://unpkg.com/maplibre-gl@3.6.2/dist/maplibre-gl.js"></script>
<link rel="stylesheet" href="https://unpkg.com/maplibre-gl@3.6.2/dist/maplibre-gl.css"/>
{% for header in headers|default([]) -%}
{{ header }}
{% endfor -%}
<body>
<div id="pymaplibregl" style="{{ style|default('height:600px;') }}"></div>
<script>
{{ js|safe }}
</script>
</body>
</html>
"""

js_template = """// ...
(() => {
    var data = {{ data|safe }};
    pymaplibregl(data);
})();
"""



from typing import Optional, Union

import geopandas as gpd

# from geopandas import read_file
from maplibre.sources import GeoJSONSource
from pydantic import BaseModel


class DataSet(BaseModel):
    url: str
    name: Optional[str] = None
    geometry_type: Union[str, list, None] = None
    bounds: Optional[tuple] = None

    @property
    def source(self) -> GeoJSONSource:
        return GeoJSONSource(data=self.url)

    def read_bounds(self) -> tuple:
        return read_file(self.url).total_bounds

    def read_data(self) -> gpd.GeoDataFrame:
        return gpd.read_file(self.url)


# TODO: Create DataSets from yaml file
class DataSets:
    vancouver_blocks: DataSet = DataSet(
        url="https://raw.githubusercontent.com/visgl/deck.gl-data/master/examples/geojson/vancouver-blocks.json",
        name="vancouver_blocks",
        geometry_type="Polygon",
        bounds=(-123.2639151, 49.1995174, -123.0234703, 49.295612),
    )

    indoor_3d_map: DataSet = DataSet(
        url="https://maplibre.org/maplibre-gl-js/docs/assets/indoor-3d-map.geojson",
        name="indoor_3d_map",
        bounds=(-87.618347, 41.865559, -87.615411, 41.866868),
    )

    urban_areas: DataSet = DataSet(
        url="https://d2ad6b4ur7yvpq.cloudfront.net/naturalearth-3.3.0/ne_50m_urban_areas.geojson",
        name="urban_areas",
        bounds=(-157.98399472, -46.26844166, 174.97002323, 69.35127106),
    )

    earthquakes: DataSet = (
        DataSet(
            url="https://docs.mapbox.com/mapbox-gl-js/assets/earthquakes.geojson",
            name="earthquakes",
            bounds=None,
        ),
    )

    bart: DataSet = DataSet(
        url="https://raw.githubusercontent.com/visgl/deck.gl-data/master/website/bart.geo.json",
        name="bart",
        geometry_type=["Point", "MultiLineString", "Polygon"],
    )



from __future__ import annotations

import itertools
from enum import Enum
from typing import Any, TypeVar

from maplibre.colors import color_brewer

T = TypeVar("T")


class GeometryType(Enum):
    POINT = "Point"
    LINE_STRING = "LineString"
    POLYGON = "Polygon"


def get_column(column: str) -> list:
    return ["get", column]


def get_geometry_type() -> list:
    return ["geometry-type"]


def geometry_type_filter(geom_type: GeometryType | str) -> list:
    return ["==", get_geometry_type(), GeometryType(geom_type).value]


def interpolate(
    property: str | list, stops: list, outputs: list, type: list = ["linear"]
) -> list:
    assert len(stops) == len(outputs)
    return [
        "interpolate",
        type,
        get_column(property) if isinstance(property, str) else property,
    ] + list(
        itertools.chain.from_iterable(
            [[stop, output] for stop, output in zip(stops, outputs)]
        )
    )


def match_expr(column: str, categories: list, outputs: list[T], fallback: T) -> list:
    assert len(categories) == len(outputs)
    return (
        ["match", get_column(column)]
        + list(
            itertools.chain.from_iterable(
                [[category, output] for category, output in zip(categories, outputs)]
            )
        )
        + [fallback]
    )


# Property examples:
# - ["zoom"],
# - ["get", "column_name"]  same as "column_name"
def step_expr(
    property: str | list,
    stops: list,
    outputs: list[T],
    fallback: T,
) -> list:
    assert len(stops) == len(outputs)
    return (
        [
            "step",
            get_column(property) if isinstance(property, str) else property,
        ]
        + list(
            itertools.chain.from_iterable(
                [[output, stop] for output, stop in zip(outputs, stops)]
            )
        )
        + [fallback]
    )


def quantile_step_expr(
    column: str, probs: list, outputs: list[T], fallback: T, values: list
) -> list:
    assert len(probs) == len(outputs)
    try:
        import numpy as np
    except ImportError as e:
        print(e)
        return []

    stops = np.quantile(values, probs)
    return step_expr(column, stops, outputs, fallback)


# -----
# ----- COLOR expression -------------------------
# -----


def color_quantile_step_expr(
    column: str, probs: list, values: Any, cmap="viridis"
) -> list:
    n = len(probs)
    colors = color_brewer(cmap, n + 1)
    return quantile_step_expr(
        column, probs, outputs=colors[0:n], fallback=colors[-1], values=values
    )


def color_step_expr(column: str, stops: list, cmap="viridis") -> list:
    n = len(stops)
    colors = color_brewer(cmap, n + 1)
    return step_expr(column, stops, outputs=colors[0:n], fallback=colors[-1])


def color_match_expr(column: str, categories: Any, cmap: str = "viridis"):
    categories = sorted(list(set(categories) - {None}))
    n = len(categories)
    colors = color_brewer(cmap, n + 1)
    return match_expr(column, categories, outputs=colors[0:n], fallback=colors[-1])


"""
def color_interpolate(
    column: str,
    stops: list,
    cmap: str = "Blues",
    colors: list = None,
    type: list = list("linear"),
) -> list:
    n = len(stops)
    colors = colors or color_brewer(cmap, n)
    return interpolate(column, stops, outputs=colors, type=type)
"""


def equal_bins_step_expr():
    pass


def filter_expr(column: str, operator: str, value: Any) -> list:
    return [operator, get_column(column), value]


def range_filter(
    column, values: tuple | list, operators: tuple | list = (">=", "<=")
) -> list:
    assert len(values) == len(operators) == 2
    return ["all"] + [[operators[i], get_column(column), values[i]] for i in range(2)]



from __future__ import annotations

import json
from enum import Enum

try:
    from pandas import DataFrame
except ImportError:
    DataFrame = None


try:
    from geopandas import GeoDataFrame
except ImportError:
    GeoDataFrame = None


class GeometryType(str, Enum):
    POINT = "Point"
    LINE_STRING = "LineString"
    POLYGON = "Polygon"


def geopandas_to_geojson(df: GeoDataFrame) -> dict:
    return json.loads(df.to_json())


def df_to_geojson(
    df: DataFrame,
    coordinates: str | list = ["lng", "lat"],
    geometry_type: str | GeometryType = GeometryType.POINT,
    properties: list = [],
) -> dict:
    geojson = {"type": "FeatureCollection", "features": []}
    for _, row in df.iterrows():
        feature = {
            "type": "Feature",
            "properties": {},
            "geometry": {"type": GeometryType(geometry_type).value, "coordinates": []},
        }
        feature["geometry"]["coordinates"] = list(row[coordinates])
        for prop in properties:
            feature["properties"][prop] = row[prop]

        geojson["features"].append(feature)

    return geojson


def get_bounds(geojson: dict) -> tuple | None:
    try:
        import shapely
    except ImportError as e:
        print(e)
        return

    return tuple(shapely.bounds(shapely.from_geojson(json.dumps(geojson))))



from __future__ import annotations

from os.path import join
from pathlib import Path

import traitlets
from anywidget import AnyWidget

from .map import Map, MapOptions


class MapWidget(AnyWidget, Map):
    """MapWidget

    Use this class to display and update maps in Jupyter Notebooks.

    See `maplibre.Map` for available methods.

    Examples:
        >>> from maplibre import MapOptions
        >>> from maplibre.ipywidget import MapWidget as Map
        >>> m = Map(MapOptions(center=(-123.13, 49.254), zoom=11, pitch=45))
        >>> m # doctest: +SKIP
    """

    _esm = join(Path(__file__).parent, "srcjs", "ipywidget.js")
    # _css = join(Path(__file__).parent, "srcjs", "maplibre-gl.css")
    _css = join(Path(__file__).parent, "srcjs", "ipywidget.css")
    _use_message_queue = True
    # _rendered = traitlets.Bool(False, config=True).tag(sync=True)
    _rendered = traitlets.Bool(False).tag(config=True).tag(sync=True)
    map_options = traitlets.Dict().tag(sync=True)
    calls = traitlets.List().tag(sync=True)
    height = traitlets.Union([traitlets.Int(), traitlets.Unicode()]).tag(sync=True)

    # Interactions Map
    clicked = traitlets.Dict().tag(sync=True)
    view_state = traitlets.Dict().tag(sync=True)

    # Interactions MapboxDraw plugin
    draw_features_selected = traitlets.List().tag(sync=True)
    draw_feature_collection_all = traitlets.Dict().tag(sync=True)

    def __init__(
        self,
        map_options=MapOptions(),
        sources: dict = None,
        layers: list = None,
        controls: list = None,
        height: int | str = 400,
        **kwargs,
    ) -> None:
        self.calls = []
        AnyWidget.__init__(self, height=height, **kwargs)
        Map.__init__(self, map_options, sources, layers, controls, **kwargs)

    """
    @traitlets.default("height")
    def _default_height(self):
        return "400px"
    """

    @traitlets.validate("height")
    def _validate_height(self, proposal):
        height = proposal["value"]
        if isinstance(height, int):
            return f"{height}px"

        return height

    @traitlets.observe("_rendered")
    def _on_rendered(self, change):
        self.send({"calls": self._message_queue, "msg": "init"})
        self._message_queue = []

    def use_message_queue(self, value: bool = True) -> None:
        self._use_message_queue = value

    def add_call(self, method_name: str, *args) -> None:
        call = [method_name, args]
        if not self._rendered:
            if not self._use_message_queue:
                self.calls = self.calls + [call]
                return

            self._message_queue.append(call)
            return

        self.send({"calls": [call], "msg": "custom call"})



# TODO: Remove file

from __future__ import annotations

from typing import Literal

from geopandas import GeoDataFrame
from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    computed_field,
    field_validator,
    model_serializer,
)

from maplibre import Layer, LayerType
from maplibre.utils import geopandas_to_geojson


class LineLayer(Layer):
    def __init__(self, source: [dict | str], id_: str = None, *args, **kwargs):
        super().__init__(LayerType.LINE, source=source, id_=id_, *args, **kwargs)


class PydanticSer(BaseModel):
    a: int = Field(..., serialization_alias="aa")
    b: int

    # @model_serializer
    def to_dict(self):
        return self.model_dump() | {"type": self.__class__.__name__}

    # @model_serializer
    def model_dump(self):
        return super().model_dump(by_alias=True) | {"type": self.__class__.__name__}


class LayoutProperties(BaseModel):
    visibility: Literal["visible", "none"] = None
    #
    fill_sort_key: list = None  # Example: ["get", "priority"]
    #
    line_cap: Literal["butt", "round", "square"] = None
    line_join: Literal["bevel", "round", "miter"] = None
    line_miter_limit: int = None
    line_round_limit: float = None
    line_sort_key: int = None


class PaintProperties(BaseModel):
    background_color: str = None
    background_emissive_strength: int = None
    background_opacity: int = None
    background_pattern: str = None


class GeoJSONifyable(object):
    def color(self, column: str):
        pass

    def color_quantile(self, column: str, q: list):
        pass

    def color_bins(self, column: str, n: int):
        pass

    def color_steps(self, column: str, steps: list):
        pass


class FeaturesCollection(object):
    def __init__(self, data):
        self._data = data

    def color(
        self,
        column: str,
        n: int = None,
        q: list = None,
        steps: list = None,
        cmap: str = None,
    ):
        return self

    def to_source(self):
        pass

    def fill(self, **kwargs):
        pass

    def line(self):
        pass


# FeaturesCollection("data").color("STATE").fill()

from typing import TypeVar, Union

from geopandas import GeoDataFrame

PandasGeoDataFrame = TypeVar("geopandas.GeoDataFrame")


class LineLayer(Layer):

    model_config = ConfigDict(arbitrary_types_allowed=True)

    data: Union[dict, GeoDataFrame]

    @field_validator("data")
    def data_validator(cls, v):
        if isinstance(v, GeoDataFrame):
            return geopandas_to_geojson(v)

        return v


def _create_layer_props(layer_type: str, props: dict) -> dict:
    return {f"{layer_type}-{k}": v for k, v in props.items() if v is not None}



from __future__ import annotations

import base64

from .map import Map

try:
    import streamlit.components.v1 as st_components
except ImportError as e:
    print(e)
    st_components = None


# TODO: Usually it should work with the 'html' component but for some reason it does not work
def st_maplibre(map: Map, height: int = 500, width: int = None) -> None:
    if st_components is None:
        return

    html_b64 = base64.b64encode(
        map.to_html(style=f"height: {height}px;").encode("utf-8")
    ).decode("utf-8")
    st_components.iframe(
        src=f"data:text/html;base64,{html_b64}",
        height=height + 16,
        scrolling=True,
        width=width,
    )



from __future__ import annotations

from pydantic import BaseModel, ConfigDict


class MapLibreBaseModel(BaseModel):
    model_config = ConfigDict(
        validate_assignment=True,
        extra="forbid",
        use_enum_values=True,
        arbitrary_types_allowed=True,
    )

    def to_dict(self) -> dict:
        return self.model_dump(by_alias=True, exclude_none=True)



from enum import Enum
from typing import Union

from pydantic import BaseModel


class DeckLayer(Enum):
    HEXAGON_LAYER = "HexagonLayer"


class PaintProperties(BaseModel):
    line_color: Union[str, list, None]
    line_opacity: Union[float, int, None]
    line_width: Union[float, int, None]
    fill_color: Union[str, list, None]
    fill_opacity: Union[float, int, None]
    fill_outline_color: Union[str, list, None]
    fill_extrusion_color: Union[str, list, None]
    fill_extrusion_opacity: Union[float, int, None]
    circle_color: Union[str, list, None]
    circle_opacity: Union[float, int, None]
    circle_stroke_color: Union[str, list, None]
    circle_radius: Union[float, int, None]



import geopandas as gpd
import maplibre.express as mx
import pandas as pd

from ..sources import GeoJSONSource, SimpleFeatures


@pd.api.extensions.register_dataframe_accessor("maplibre")
class MapLibreAccessor(object):
    def __init__(self, gdf: gpd.GeoDataFrame):
        self._gdf = gdf

    def to_source(self) -> GeoJSONSource:
        return SimpleFeatures(self._gdf).to_source()

    def fill(self) -> mx.SimpleLayer:
        return mx.fill(self._gdf)



from __future__ import annotations

import json
import webbrowser
from typing import Union

from jinja2 import Template
from pydantic import ConfigDict, Field, field_validator

from ._core import MapLibreBaseModel
from ._templates import html_template, js_template
from ._utils import get_temp_filename, read_internal_file
from .basemaps import (Carto, MapTiler, construct_carto_basemap_url,
                       construct_maptiler_basemap_url)
from .controls import Control, ControlPosition, Marker
from .layer import Layer
from .plugins import MapboxDrawOptions
from .sources import SimpleFeatures, Source

try:
    from geopandas import GeoDataFrame
except ImportError:
    GeoDataFrame = None

try:
    import pydeck
except ImportError:
    pydeck = None


def parse_deck_layers(layers: list[dict | "pydeck.Layer"]) -> list[dict]:
    for i, layer in enumerate(layers):
        if pydeck is not None and isinstance(layer, pydeck.Layer):
            layers[i] = json.loads(layer.to_json())

    return layers


class MapOptions(MapLibreBaseModel):
    """Map options

    Note:
        See [MapOptions](https://maplibre.org/maplibre-gl-js/docs/API/type-aliases/MapOptions) for more details.
    """

    model_config = ConfigDict(
        validate_assignment=True, extra="forbid", use_enum_values=False
    )
    antialias: bool = None
    attribution_control: bool = Field(None, serialization_alias="attributionControl")
    bearing: Union[int, float] = None
    bearing_snap: int = Field(None, serialization_alias="bearingSnap")
    bounds: tuple = None
    box_zoom: bool = Field(None, serialization_alias="boxZoom")
    center: tuple = None
    click_tolerance: int = Field(None, serialization_alias="clickTolerance")
    custom_attribution: bool = Field(None, serialization_alias="customAttribution")
    double_click_zoom: bool = Field(None, serialization_alias="doubleClickZoom")
    fade_duration: int = Field(None, serialization_alias="fadeDuration")
    fit_bounds_options: dict = Field(None, serialization_alias="fitBoundsOptions")
    hash: Union[bool, str] = None
    interactive: bool = None
    keyword: bool = None
    max_bounds: tuple = Field(None, serialization_alias="maxBounds")
    max_pitch: int = Field(None, serialization_alias="maxPitch")
    max_zoom: int = Field(None, serialization_alias="maxZoom")
    min_pitch: int = Field(None, serialization_alias="minPitch")
    min_zoom: int = Field(None, serialization_alias="minZoom")
    pitch: Union[int, float] = None
    scroll_zoom: bool = Field(None, serialization_alias="scrollZoom")
    style: Union[str, Carto, MapTiler, dict] = construct_carto_basemap_url(
        Carto.DARK_MATTER
    )
    zoom: Union[int, float] = None

    @field_validator("style")
    def validate_style(cls, v):
        if isinstance(v, Carto):
            return construct_carto_basemap_url(v)

        if isinstance(v, MapTiler):
            return construct_maptiler_basemap_url(v)

        return v


class Map(object):
    """Map

    Args:
        map_options (MapOptions): Map options.
        sources (dict): Sources to be added to the map. Keys are source IDs.
        layers (list): Layers to be added to the map.
        controls (list): Controls to be added to the map.
        **kwargs: Keyword arguments that are appended to the `MapOptions` object.

    Examples:
        >>> from maplibre.map import Map, MapOptions
        >>> map_options = MapOptions(center=(9.5, 51.31667), zoom=8)
        >>> m = Map(map_options)
        >>> dict(m)
        {'mapOptions': {'center': (9.5, 51.31667), 'style': 'https://basemaps.cartocdn.com/gl/dark-matter-gl-style/style.json', 'zoom': 8}, 'calls': []}
    """

    MESSAGE = "not implemented yet"

    def __init__(
        self,
        map_options: MapOptions = MapOptions(),
        sources: dict = None,
        layers: list = None,
        controls: list = None,
        **kwargs,
    ):
        self.map_options = (
            map_options.to_dict() | kwargs
        )  # MapOptions(**kwargs).to_dict() # need to fix MapWidget, because height is passed as kwarg
        self._message_queue = []
        self.add_layers(layers, sources)
        if controls:
            for control in controls:
                self.add_control(control)

    def __iter__(self):
        for k, v in self.to_dict().items():
            yield k, v

    def to_dict(self) -> dict:
        return {"mapOptions": self.map_options, "calls": self._message_queue}

    """
    @property
    def sources(self) -> list:
        return [item["data"] for item in self._calls if item["name"] == "addSource"]

    @property
    def layers(self) -> list:
        return [item["data"] for item in self._calls if item["name"] == "addLayer"]
    """

    # TODO: Rename to add_map_call
    """
    def add_call_(self, func_name: str, params: list) -> None:
        self._message_queue.append(
            {"name": "applyFunc", "data": {"funcName": func_name, "params": params}}
        )
    """

    def add_call(self, method_name: str, *args) -> None:
        """Add a method call that is executed on the map instance

        Args:
            method_name (str): The name of the map method to be executed.
            *args (any): The arguments to be passed to the map method.
        """
        # TODO: Pass as dict? {"name": method_name, "args": args}
        call = [method_name, args]
        self._message_queue.append(call)

    def add_control(
        self,
        control: Control,
        position: [str | ControlPosition] = None,
    ) -> None:
        """Add a control to the map

        Args:
            control (Control): The control to be added to the map.
            position (str | ControlPosition): The position of the control.
        """
        position = position or control.position
        self.add_call(
            "addControl",
            control.type,
            control.to_dict(),
            ControlPosition(position).value,
        )

    def add_source(self, id: str, source: [Source | dict | GeoDataFrame]) -> None:
        """Add a source to the map

        Args:
            id (str): The unique ID of the source.
            source (Source | dict | GeoDataFrame): The source to be added to the map.
        """
        if GeoDataFrame is not None and isinstance(source, GeoDataFrame):
            source = SimpleFeatures(source).to_source()

        if isinstance(source, Source):
            source = source.to_dict()

        self.add_call("addSource", id, source)

    def add_layer(self, layer: [Layer | dict], before_id: str = None) -> None:
        """Add a layer to the map

        Args:
            layer (Layer | dict): The Layer to be added to the map.
            before_id (str): The ID of an existing layer to insert the new layer before,
                resulting in the new layer appearing visually beneath the existing layer.
                If `None`, the new layer will appear above all other layers.
        """
        if isinstance(layer, Layer):
            layer = layer.to_dict()

        self.add_call("addLayer", layer, before_id)

    def add_layers(self, layers: list = None, sources: dict = None):
        layers = layers or []
        sources = sources or dict()
        for source_id, source in sources.items():
            self.add_source(source_id, source)

        for layer in layers:
            self.add_layer(layer)

    def add_marker(self, marker: Marker) -> None:
        """Add a marker to the map

        Args:
            marker (Marker): The marker to be added to the map.
        """
        self.add_call("addMarker", marker.to_dict())

    def add_popup(self, layer_id: str, prop: str = None, template: str = None) -> None:
        """Add a popup to the map

        Args:
            layer_id (str): The layer to which the popup is added.
            prop (str): The property of the source to be displayed. If `None`, all properties are displayed.
            template (str): A mustache template. If supplied, `prop` is ignored.
        """
        self.add_call("addPopup", layer_id, prop, template)

    def add_tooltip(
        self, layer_id: str, prop: str = None, template: str = None
    ) -> None:
        """Add a tooltip to the map

        Args:
            layer_id (str): The layer to which the tooltip is added.
            prop (str): The property of the source to be displayed. If `None`, all properties are displayed.
            template (str): A mustache template. If supplied, `prop` is ignored.

        Examples:
            >>> m = Map()
            >>> # ...
            >>> m.add_tooltip("test-layer", template="Name: {{ name }}")
        """
        self.add_call("addTooltip", layer_id, prop, template)

    def set_filter(self, layer_id: str, filter_: list):
        """Update the filter of a layer

        Args:
            layer_id (str): The name of the layer to be updated.
            filter_ (list): The filter expression that is applied to the source of the layer.
        """
        self.add_call("setFilter", layer_id, filter_)

    def set_paint_property(self, layer_id: str, prop: str, value: any) -> None:
        """Update the paint property of a layer

        Args:
            layer_id (str): The name of the layer to be updated.
            prop (str): The name of the paint property to be updated.
            value (any): The new value of the paint property.
        """
        self.add_call("setPaintProperty", layer_id, prop, value)

    def set_layout_property(self, layer_id: str, prop: str, value: any) -> None:
        """Update a layout property of a layer

        Args:
            layer_id (str): The name of the layer to be updated.
            prop (str): The name of the layout property to be updated.
            value (any): The new value of the layout property.
        """
        self.add_call("setLayoutProperty", layer_id, prop, value)

    def set_data(self, source_id: str, data: dict | GeoDataFrame) -> None:
        """Update the data of a GeoJSON source

        Args:
            source_id (str): The name of the source to be updated.
            data (dict): The data of the source.
        """
        if isinstance(data, GeoDataFrame):
            data = SimpleFeatures(data).to_source().data

        self.add_call("setSourceData", source_id, data)

    def set_visibility(self, layer_id: str, visible: bool = True) -> None:
        """Update the visibility of a layer

        Args:
            layer_id (str): The name of the layer to be updated.
            visible (bool): Whether the layer is visible or not.
        """
        value = "visible" if visible else "none"
        self.add_call("setLayoutProperty", layer_id, "visibility", value)

    def fit_bounds(
        self,
        bounds: tuple | list = None,
        data: GeoDataFrame = None,
        animate=False,
        **kwargs,
    ) -> None:
        """Pan and zoom the map to contain its visible area within the specified geographical bounds"""
        kwargs["animate"] = animate
        if data is not None:
            bounds = tuple(data.total_bounds)

        self.add_call("fitBounds", bounds, kwargs)

    def to_html(self, title: str = "My Awesome Map", **kwargs) -> str:
        """Render to html

        Args:
            title (str): The Title of the HTML document.
            **kwargs (Any): Additional keyword arguments that are passed to the template.
                Currently, `style` is the only supported keyword argument.

        Examples:
            >>> from maplibre import Map
            >>> m = Map()
            >>> with open("/tmp/map.html", "w") as f:
            ...     f.write(m.to_html(style="height: 800px;") # doctest: +SKIP
        """
        js_lib = read_internal_file("srcjs", "pywidget.js")
        js_snippet = Template(js_template).render(data=json.dumps(self.to_dict()))
        css = read_internal_file("srcjs", "pywidget.css")
        headers = [f"<style>{css}</style>"]

        # Deck.GL headers
        add_deckgl_headers = "addDeckOverlay" in [
            item[0] for item in self._message_queue
        ]
        # TODO: Set version in constants
        deckgl_headers = (
            [
                # '<script src="https://unpkg.com/h3-js"></script>',
                '<script src="https://unpkg.com/h3-js@4.1.0/dist/h3-js.umd.js"></script>',
                '<script src="https://unpkg.com/deck.gl@9.0.16/dist.min.js"></script>',
                '<script src="https://unpkg.com/@deck.gl/json@9.0.16/dist.min.js"></script>',
            ]
            if add_deckgl_headers
            else []
        )

        # Mapbox Draw headers
        add_mapbox_draw_headers = "addMapboxDraw" in [
            item[0] for item in self._message_queue
        ]
        # TODO: Set version in constants
        mapbox_draw_headers = (
            [
                # "<script src='https://api.mapbox.com/mapbox-gl-js/plugins/mapbox-gl-draw/v1.4.3/mapbox-gl-draw.js'></script>",
                # "<link rel='stylesheet' href='https://api.mapbox.com/mapbox-gl-js/plugins/mapbox-gl-draw/v1.4.3/mapbox-gl-draw.css' type='text/css' />",
                "<script src='https://www.unpkg.com/@mapbox/mapbox-gl-draw@1.4.3/dist/mapbox-gl-draw.js'></script>",
                "<link rel='stylesheet' href='https://www.unpkg.com/@mapbox/mapbox-gl-draw@1.4.3/dist/mapbox-gl-draw.css' type='text/css' />",
            ]
            if add_mapbox_draw_headers
            else []
        )

        output = Template(html_template).render(
            js="\n".join([js_lib, js_snippet]),
            title=title,
            headers=headers + deckgl_headers + mapbox_draw_headers,
            **kwargs,
        )
        return output

    def save(self, filename: str = None, preview=True, **kwargs):
        """Save the map to an HTML file"""
        return save_map(self, filename, preview, **kwargs)

    # -------------------------
    # Plugins
    # -------------------------
    def add_deck_layers(
        self, layers: list[dict | "pydeck.Layer"], tooltip: str | dict = None
    ) -> None:
        """Add Deck.GL layers to the layer stack

        Args:
            layers (list[dict | "pydeck.Layer"]): A list of dictionaries containing the Deck.GL layers to be added.
            tooltip (str | dict): Either a single mustache template string applied to all layers
                or a dictionary where keys are layer ids and values are mustache template strings.
        """
        layers = parse_deck_layers(layers)
        self.add_call("addDeckOverlay", layers, tooltip)

    def set_deck_layers(
        self, layers: list[dict | "pydeck.Layer"], tooltip: str | dict = None
    ) -> None:
        """Update Deck.GL layers

        Args:
            layers (list[dict | "pydeck.Layer"]): A list of dictionaries containing the Deck.GL layers to be updated.
                New layers will be added. Missing layers will be removed.
            tooltip (str | dict): Must be set to keep tooltip even if it did not change.
        """
        layers = parse_deck_layers(layers)
        self.add_call("setDeckLayers", layers, tooltip)

    def add_mapbox_draw(
        self,
        options: dict | MapboxDrawOptions = None,
        position: str | ControlPosition = ControlPosition.TOP_LEFT,
        geojson: dict = None,
    ) -> None:
        """Add MapboxDraw controls to the map

        Note:
            See [MapboxDraw API Reference](https://github.com/mapbox/mapbox-gl-draw/blob/main/docs/API.md)
                for available options.

        Args:
            options (dict | MapboxDrawOptions): MapboxDraw options.
            position (str | ControlPosition): The position of the MapboxDraw controls.
            geojson (dict): A GeoJSON Feature, FeatureCollection or Geometry to be added to the draw layer.
        """
        if isinstance(options, MapboxDrawOptions):
            options = options.to_dict()

        self.add_call(
            "addMapboxDraw", options or {}, ControlPosition(position).value, geojson
        )


def save_map(m: Map, filename: str = None, preview=True, **kwargs) -> str:
    if not filename:
        filename = get_temp_filename()

    with open(filename, "w") as f:
        f.write(m.to_html(**kwargs))

    if preview:
        webbrowser.open(filename)

    return filename



from __future__ import annotations

from typing import Union

from pydantic import Field
from typing_extensions import Self

from maplibre.colors import color_brewer
from maplibre.config import options
from maplibre.controls import NavigationControl
from maplibre.expressions import (
    GeometryType,
    color_match_expr,
    color_quantile_step_expr,
    color_step_expr,
    geometry_type_filter,
    interpolate,
)
from maplibre.layer import Layer, LayerType
from maplibre.map import Map, MapOptions
from maplibre.sources import GeoJSONSource, SimpleFeatures

try:
    import geopandas as gpd
except ImportError as e:
    print(e)
    gpd = None


def path_is_geojson_url(path: str) -> bool:
    if path is None:
        return False

    path = path.lower()
    return path.startswith("http") and path.endswith("json")


# TODO: Maybe rename to DataLayer?
class SimpleLayer(Layer):
    sf: Union[SimpleFeatures, gpd.GeoDataFrame, str] = Field(exclude=True)

    def model_post_init(self, __context) -> None:
        if not isinstance(self.sf, SimpleFeatures):
            self.sf = SimpleFeatures(self.sf)

        sf_path = self.sf.path
        if path_is_geojson_url(sf_path):
            self.source = GeoJSONSource(data=sf_path)
        else:
            self.source = self.sf.to_source()

        # TODO: Use layer paint properties from options
        if self.paint is None:
            layer_type = LayerType(self.type).value
            self.paint = {f"{layer_type}-color": options.fallback_color}

    def _set_paint_property(self, prop, value):
        layer_type = LayerType(self.type).value
        self.paint[f"{layer_type}-{prop}"] = value

    def color(self, value: str | list) -> Self:
        self._set_paint_property("color", value)
        return self

    def opacity(self, value: float) -> Self:
        self._set_paint_property("opacity", value)
        return self

    def color_category(self, column: str, cmap: str = options.cmap) -> Self:
        expr = color_match_expr(column, categories=self.sf.data[column], cmap=cmap)
        self._set_paint_property("color", expr)
        return self

    def color_quantile(
        self,
        column: str,
        probs: list = [0.1, 0.25, 0.5, 0.75],
        cmap: str = options.cmap,
    ) -> Self:
        expr = color_quantile_step_expr(
            column, probs, values=self.sf.data[column], cmap=cmap
        )
        self._set_paint_property("color", expr)
        return self

    def color_bin(
        self, column: str, stops: list = None, n: int = None, cmap=options.cmap
    ) -> Self:
        if stops is None and n is None:
            pass

        expr = color_step_expr(column, stops, cmap)
        self._set_paint_property("color", expr)
        return self

    def interpolate_color(
        self, column: str, stops=None, colors=("yellow", "red")
    ) -> Self:
        stops = stops or [f(self.sf.data[column]) for f in [min, max]]
        expr = interpolate(column, stops, colors)
        self._set_paint_property("color", expr)
        return self

    def to_map(
        self,
        map_options: MapOptions = MapOptions(),
        controls: list = None,
        tooltip: bool = True,
        fit_bounds: bool = True,
        **kwargs,
    ) -> Map:
        controls = controls or [NavigationControl()]
        if fit_bounds:
            map_options.bounds = self.sf.bounds

        m = Map(map_options, layers=[self], controls=controls, **kwargs)
        if tooltip:
            m.add_tooltip(self.id)

        return m


def _create_prop_key(layer_type: str, prop: str) -> str:
    return "-".join([layer_type, prop])


def fill(data: gpd.GeoDataFrame | str, **kwargs) -> SimpleLayer:
    if "paint" not in kwargs:
        kwargs["paint"] = options.paint_props[LayerType.FILL.value]

    return SimpleLayer(
        type=LayerType.FILL,
        sf=data,
        **kwargs,
    )


def circle(data: gpd.GeoDataFrame | str, **kwargs) -> SimpleLayer:
    if "paint" not in kwargs:
        kwargs["paint"] = options.paint_props[LayerType.CIRCLE.value]

    return SimpleLayer(
        type=LayerType.CIRCLE,
        sf=data,
        **kwargs,
    )


def line(data: gpd.GeoDataFrame | str, **kwargs) -> SimpleLayer:
    if "paint" not in kwargs:
        kwargs["paint"] = options.paint_props[LayerType.LINE.value]

    return SimpleLayer(
        type=LayerType.LINE,
        sf=data,
        **kwargs,
    )


def fill_extrusion(
    data: gpd.GeoDataFrame | str,
    fill_extrusion_base: int | float | list = None,
    fill_extrusion_height: int | float | list = None,
    **kwargs,
) -> SimpleLayer:
    if "paint" not in kwargs:
        kwargs["paint"] = options.paint_props[
            LayerType.FILL_EXTRUSION.value.replace("-", "_")
        ]

    if fill_extrusion_base is not None:
        kwargs["paint"]["fill-extrusion-base"] = fill_extrusion_base

    if fill_extrusion_height is not None:
        kwargs["paint"]["fill-extrusion-height"] = fill_extrusion_height

    return SimpleLayer(type=LayerType.FILL_EXTRUSION, sf=data, **kwargs)


# TODO: Use default layers from settings
def fill_line_circle(source_id: str, colors: list = None) -> list:
    if colors is not None:
        assert len(colors) == 3
    else:
        colors = color_brewer(options.cmap, 3)

    fill_color, line_color, circle_color = colors

    fill_layer = Layer(
        type=LayerType.FILL,
        source=source_id,
        filter=geometry_type_filter(GeometryType.POLYGON),
    ).set_paint_props(
        fill_color=fill_color, fill_outline_color=options.fill_outline_color
    )

    line_layer = Layer(
        type=LayerType.LINE,
        source=source_id,
        filter=geometry_type_filter(GeometryType.LINE_STRING),
    ).set_paint_props(line_color=line_color)

    circle_layer = Layer(
        type=LayerType.CIRCLE,
        source=source_id,
        filter=geometry_type_filter(GeometryType.POINT),
    ).set_paint_props(circle_color=circle_color)

    return [fill_layer, line_layer, circle_layer]


def map_this(data: gpd.GeoDataFrame | str, tooltip: bool = True, **kwargs) -> Map:
    sf = SimpleFeatures(data)
    layers = fill_line_circle(sf.source_id)
    kwargs["bounds"] = sf.bounds
    map_options = MapOptions(**kwargs)
    m = Map(
        map_options,
        sources=sf.to_sources_dict(),
        layers=layers,
        controls=[NavigationControl()],
    )
    if tooltip:
        for layer in layers:
            m.add_tooltip(layer.id)

    return m



__version__ = "0.2.6"
_shiny_output_class = "shiny-maplibregl-output"

# TODO: Remove from here
# tmp
_points = "https://raw.githubusercontent.com/visgl/deck.gl-data/master/examples/scatterplot/manhattan.json"
_geojson = "https://docs.mapbox.com/mapbox-gl-js/assets/indoor-3d-map.geojson"
_urban_areas = "https://d2ad6b4ur7yvpq.cloudfront.net/naturalearth-3.3.0/ne_50m_urban_areas.geojson"
_highway_lines = (
    "https://github.com/visgl/deck.gl-data/raw/master/examples/highway/roads.json"
)

_airports = (
    "https://github.com/visgl/deck.gl-data/raw/master/examples/line/airports.json"
)
_flights = "https://github.com/visgl/deck.gl-data/raw/master/examples/line/heathrow-flights.json"



from .config import options
from .controls import ControlPosition, ControlType
from .layer import Layer, LayerType
from .map import Map, MapOptions
from .mapcontext import MapContext

# TODO: Only import once, preferred: MapLibreRenderer
from .renderer import MapLibreRenderer
from .renderer import MapLibreRenderer as render_maplibregl
from .ui import output_maplibregl



from __future__ import annotations

from enum import Enum

from .config import options
from .layer import Layer, LayerType

MAPLIBRE_DEMO_TILES = "https://demotiles.maplibre.org/style.json"


class Carto(Enum):
    """Carto basemap styles

    Attributes:
        DARK_MATTER: dark-matter
        POSITRON: positron
        VOYAGER: voyager
        POSITRON_NOLABELS: positron-nolabels
        DARK_MATTER_NOLABELS: dark-matter-nolabels
        VOYAGER_NOLABELS: voyager-nolabels

    Examples:
        >>> from maplibre import Map, MapOptions
        >>> from maplibre.basemaps import Carto

        >>> m = Map(MapOptions(style=Carto.DARK_MATTER))
    """

    DARK_MATTER = "dark-matter"
    POSITRON = "positron"
    VOYAGER = "voyager"
    POSITRON_NOLABELS = "positron-nolabels"
    DARK_MATTER_NOLABELS = "dark-matter-nolabels"
    VOYAGER_NOLABELS = "voyager-nolabels"


def construct_carto_basemap_url(style_name: str | Carto = "dark-matter") -> str:
    return f"https://basemaps.cartocdn.com/gl/{Carto(style_name).value}-gl-style/style.json"


def construct_basemap_style(
    name: str = "nice-style", sources: dict = {}, layers: list = []
) -> dict:
    """Construct a basemap style

    Args:
        name (str): The name of the basemap style.
        sources (dict): The sources to be used for the basemap style.
        layers (list): The layers to be used for the basemap style.
    """
    layers = [
        layer.to_dict() if isinstance(layer, Layer) else layer for layer in layers
    ]
    return {"name": name, "version": 8, "sources": sources, "layers": layers}


def background(color: str = "black", opacity: float = 1.0) -> dict:
    bg_layer = Layer(
        type=LayerType.BACKGROUND,
        id="background",
        source=None,
        paint={"background-color": color, "background-opacity": opacity},
    )
    return construct_basemap_style(layers=[bg_layer])


class MapTiler(Enum):
    AQUARELLE = "aquarelle"
    BACKDROP = "backdrop"
    BASIC = "basic"
    BRIGHT = "bright"
    DATAVIZ = "dataviz"
    LANDSCAPE = "landscape"
    OCEAN = "ocean"
    OPEN_STREET_MAP = "openstreetmap"
    OUTDOOR = "outdoor"
    SATELLITE = "satellite"
    STREETS = "streets"
    TONER = "toner"
    TOPO = "topo"
    WINTER = "winter"


def construct_maptiler_basemap_url(style_name: str | MapTiler = "aquarelle") -> str:
    maptiler_api_key = options.maptiler_api_key
    if isinstance(style_name, MapTiler):
        style_name = MapTiler(style_name).value

    return (
        f"https://api.maptiler.com/maps/{style_name}/style.json?key={maptiler_api_key}"
    )



from __future__ import annotations

import os
import pathlib
from tempfile import mkdtemp
from uuid import uuid4


def fix_keys(d: dict) -> dict:
    return {k.replace("_", "-"): v for k, v in d.items() if v is not None}


def get_output_dir(output_dir: str = None, prefix: str = "py-maplibre-gl-") -> str:
    if not output_dir:
        output_dir = mkdtemp(prefix=prefix)
    else:
        pathlib.Path(output_dir).mkdir(parents=True, exist_ok=True)

    return output_dir


def get_temp_filename(file_extension: str = ".html") -> str:
    return get_output_dir() + os.sep + str(uuid4()).replace("-", "") + file_extension


def get_internal_file_path(*args):
    # print(os.path.dirname(__file__))
    return os.path.join(os.path.dirname(__file__), *args)


def read_internal_file(*args):
    with open(get_internal_file_path(*args)) as f:
        content = f.read()

    return content



from shiny.session import Session, require_active_session

from .map import Map


class MapContext(Map):
    """MapContext

    Use this class to update a `Map` instance in a Shiny app.
    Must be used inside an async function.

    See `maplibre.Map` for available methods.

    Args:
        id (string): The id of the map to be updated.
        session (Session): A Shiny session.
            If `None`, the active session is used.
    """

    def __init__(self, id: str, session: Session = None) -> None:
        self.id = id
        self._session = require_active_session(session)
        self.map_options = {}
        self._message_queue = []

    async def __aenter__(self):
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        await self.render()

    async def render(self):
        await self._session.send_custom_message(
            f"pymaplibregl-{self.id}", {"id": self.id, "calls": self._message_queue}
        )
